#!/usr/bin/env python3
"""
Main steering script for ntuple injection
"""
import argparse
import os
import gc
from multiprocessing import Pool
import tensorflow as tf
from HelperModules.Injector import Injector
from HelperModules.MessageHandler import InjectorMessage, InjectorDoneMessage, WelcomeMessage, EnvironmentalCheck

def inject(args, fname, count, max_count):
    """
    Main function to inject ntuples in a parallel way
    """
    # Here we create the necessary settings objects and read information from the config file
    InjectorMessage("Processing "+str(count)+"/"+str(max_count)+": "+fname)
    Inj = Injector(args)
    Inj.do_Injection(fname)
    InjectorDoneMessage("Finished "+fname)
    del Inj
    gc.collect()
    
def absoluteFilePaths(directory):
    """
    Method to get the absolute file paths w.r.t. a given directory path
    """
    for dirpath, _, filenames in os.walk(directory):
        for f in filenames:
            yield os.path.abspath(os.path.join(dirpath, f))

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    PARSER.add_argument("-o", "--output_path", help="output path for the converted ntuples", required=True)
    PARSER.add_argument("-c", "--configfile", help="Config file", nargs='+', required=True)
    PARSER.add_argument("-i", "--input_path", help="Ntuple path", required=True)
    PARSER.add_argument("-f", "--filter", dest="wildcard", help="additional string to filter input files", type=str, default='')
    PARSER.add_argument("-t", "--treefilter", dest="tree_wildcard", help="additional string to filter trees", type=str, default='')
    PARSER.add_argument("-m", "--model-path", help="input path of the ML model and history", nargs='+', required=True)
    PARSER.add_argument("--processes", help="the number of parallel processes to run", type=int, default=8)
    ARGS = PARSER.parse_args()

    #Ignoring tf warnings and info messages
    tf.get_logger().setLevel('ERROR')

    WelcomeMessage("Injector")
    EnvironmentalCheck(check="env")

    # Let's make sure the output directory exists. This is done recursively and should work for python>3.5
    for dirpath, dirnames, filenames in os.walk(ARGS.input_path):
        structure = os.path.join(ARGS.output_path, dirpath[len(ARGS.input_path):].strip("/"))
        if not os.path.isdir(structure):
            os.makedirs(structure)
    fnames = absoluteFilePaths(ARGS.input_path)
    fnames = [f for f in fnames if ARGS.wildcard in f or ARGS.wildcard=='']
    p = Pool(ARGS.processes, maxtasksperchild=1)
    MAX_ID = len(fnames)
    # for i,f in enumerate(fnames):
    #     inject(ARGS,f,i+1,MAX_ID)
    results = [p.apply_async(inject, (ARGS, f, i+1, MAX_ID,)) for i,f in enumerate(fnames)]
    results_g = [r.get() for r in results]
    p.close()
    p.join()
